package ru.dbotthepony.kstarbound.json

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet

fun <T> TypeAdapter<T>.transformRead(transformer: (T) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@transformRead.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			return transformer(this@transformRead.read(`in`))
		}
	}
}

fun <T> TypeAdapter<T>.transformWrite(transformer: (T) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@transformWrite.write(out, transformer(value))
		}

		override fun read(`in`: JsonReader): T {
			return this@transformWrite.read(`in`)
		}
	}
}

fun <In, Out> TypeAdapter<In>.transform(read: (In) -> Out, write: (Out) -> In): TypeAdapter<Out> {
	return object : TypeAdapter<Out>() {
		override fun write(out: JsonWriter, value: Out) {
			return this@transform.write(out, write(value))
		}

		override fun read(`in`: JsonReader): Out {
			return read(this@transform.read(`in`))
		}
	}
}

fun <T> TypeAdapter<T>.ifString(reader: (String) -> T): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			return this@ifString.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			if (`in`.peek() == JsonToken.STRING) {
				return reader(`in`.nextString())
			}

			return this@ifString.read(`in`)
		}
	}
}

fun <T> TypeAdapter<T?>.neverNull(): TypeAdapter<T> {
	return object : TypeAdapter<T>() {
		override fun write(out: JsonWriter, value: T) {
			this@neverNull.write(out, value)
		}

		override fun read(`in`: JsonReader): T {
			val path = `in`.path
			return this@neverNull.read(`in`) ?: throw JsonSyntaxException("Value was null near $path")
		}
	}
}

fun <T> TypeAdapter<T>.allowNull(): TypeAdapter<T?> = nullSafe()

@Deprecated("Небезопасно, надо использовать другой путь")
fun TypeAdapterFactory.ifString(reader: (String) -> Any): TypeAdapterFactory {
	return object : TypeAdapterFactory {
		override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			return this@ifString.create(gson, type)?.ifString(reader as (String) -> T)
		}
	}
}

fun JsonReader.consumeNull(): Boolean {
	if (peek() == JsonToken.NULL) {
		nextNull()
		return true
	}

	return false
}

fun JsonWriter.value(element: JsonPrimitive) {
	if (element.isBoolean) {
		value(element.asBoolean)
	} else if (element.isNumber) {
		value(element.asNumber)
	} else if (element.isString) {
		value(element.asString)
	} else {
		throw IllegalArgumentException(element.toString())
	}
}

fun JsonWriter.value(element: JsonNull) {
	nullValue()
}

fun JsonWriter.value(element: JsonArray) {
	beginArray()
	for (v in element) value(v)
	endArray()
}

fun JsonWriter.value(element: JsonObject) {
	beginObject()
	for ((k, v) in element.entrySet()) {
		name(k)
		value(v)
	}
	endObject()
}

fun JsonWriter.value(element: JsonElement?) {
	when (element) {
		is JsonPrimitive -> value(element)
		is JsonNull -> value(element)
		is JsonArray -> value(element)
		is JsonObject -> value(element)
		null -> nullValue()
		else -> throw IllegalArgumentException(element.toString())
	}
}

inline fun <reified C : Collection<E>, reified E> Gson.collectionAdapter(): TypeAdapter<C> {
	return getAdapter(TypeToken.getParameterized(C::class.java, E::class.java)) as TypeAdapter<C>
}

inline fun <reified E> Gson.listAdapter(): TypeAdapter<ImmutableList<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.mutableListAdapter(): TypeAdapter<ArrayList<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.setAdapter(): TypeAdapter<ImmutableSet<E>> {
	return collectionAdapter()
}

inline fun <reified E> Gson.mutableSetAdapter(): TypeAdapter<ObjectOpenHashSet<E>> {
	return collectionAdapter()
}

fun JsonArray(elements: Collection<JsonElement>): JsonArray {
	return JsonArray(elements.size).also { elements.forEach(it::add) }
}

fun JsonArray.clear() {
	while (size() > 0) {
		remove(size() - 1)
	}
}

operator fun JsonObject.set(key: String, value: JsonElement?) { add(key, value) }
operator fun JsonObject.set(key: String, value: String) { add(key, JsonPrimitive(value)) }
operator fun JsonObject.set(key: String, value: Int) { add(key, JsonPrimitive(value)) }
operator fun JsonObject.set(key: String, value: Long) { add(key, JsonPrimitive(value)) }
operator fun JsonObject.set(key: String, value: Float) { add(key, JsonPrimitive(value)) }
operator fun JsonObject.set(key: String, value: Double) { add(key, JsonPrimitive(value)) }
operator fun JsonObject.set(key: String, value: Boolean) { add(key, InternedJsonElementAdapter.of(value)) }
operator fun JsonObject.set(key: String, value: Nothing?) { add(key, JsonNull.INSTANCE) }
operator fun JsonObject.contains(key: String): Boolean = has(key)

fun JsonArray.get(key: Int, default: Boolean): Boolean {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isBoolean) {
		throw JsonSyntaxException("Expected boolean at $key; got $value")
	}

	return value.asBoolean
}

fun JsonObject.get(key: String, default: Boolean): Boolean {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isBoolean) {
		throw JsonSyntaxException("Expected boolean at $key; got $value")
	}

	return value.asBoolean
}

fun JsonObject.get(key: String, default: String): String {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isString) {
		throw JsonSyntaxException("Expected string at $key; got $value")
	}

	return value.asString
}

fun JsonArray.get(key: Int, default: String): String {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isString) {
		throw JsonSyntaxException("Expected string at $key; got $value")
	}

	return value.asString
}

fun JsonObject.get(key: String, default: Int): Int {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected integer at $key; got $value")
	}

	return value.asInt
}

fun JsonArray.get(key: Int, default: Int): Int {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected integer at $key; got $value")
	}

	return value.asInt
}

fun JsonObject.get(key: String, default: Long): Long {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected long at $key; got $value")
	}

	return value.asLong
}

fun JsonArray.get(key: Int, default: Long): Long {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected long at $key; got $value")
	}

	return value.asLong
}

fun JsonObject.get(key: String, default: Double): Double {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected double at $key; got $value")
	}

	return value.asDouble
}

fun JsonArray.get(key: Int, default: Double): Double {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonPrimitive || !value.isNumber) {
		throw JsonSyntaxException("Expected double at $key; got $value")
	}

	return value.asDouble
}

fun JsonObject.get(key: String, default: JsonObject): JsonObject {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonObject)
		throw JsonSyntaxException("Expected json object at $key; got $value")

	return value
}

inline fun <reified T : JsonElement> JsonObject.get(key: String, default: () -> T): T {
	if (!has(key))
		return default.invoke()

	val value = this[key]

	if (value !is T)
		throw JsonSyntaxException("Expected json object at $key; got $value")

	return value
}

fun JsonArray.get(key: Int, default: JsonObject): JsonObject {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonObject)
		throw JsonSyntaxException("Expected json object at $key; got $value")

	return value
}

inline fun JsonArray.get(key: Int, default: () -> JsonObject): JsonObject {
	if (key !in 0 until size())
		return default.invoke()

	val value = this[key]

	if (value !is JsonObject)
		throw JsonSyntaxException("Expected json object at $key; got $value")

	return value
}

fun JsonObject.get(key: String, default: JsonArray): JsonArray {
	if (!has(key))
		return default

	val value = this[key]

	if (value !is JsonArray)
		throw JsonSyntaxException("Expected json array at $key; got $value")

	return value
}

fun JsonArray.get(key: Int, default: JsonArray): JsonArray {
	if (key !in 0 until size())
		return default

	val value = this[key]

	if (value !is JsonArray)
		throw JsonSyntaxException("Expected json array at $key; got $value")

	return value
}

inline fun <reified T : JsonElement> JsonArray.get(key: Int, default: () -> T): T {
	if (key !in 0 until size())
		return default.invoke()

	val value = this[key]

	if (value !is T)
		throw JsonSyntaxException("Expected json array at $key; got $value")

	return value
}

fun JsonObject.get(key: String, default: JsonPrimitive): JsonElement {
	if (!has(key))
		return default

	return this[key]
}

fun JsonArray.get(key: Int, default: JsonPrimitive): JsonElement {
	if (key !in 0 until size())
		return default

	return this[key]
}

fun <T> JsonObject.get(key: String, type: TypeAdapter<T>): T {
	return get(key, type) { throw JsonSyntaxException("Expected value at $key, got nothing") }
}

inline fun <T> JsonObject.get(key: String, type: TypeAdapter<out T>, orElse: () -> T): T {
	if (!has(key))
		return orElse.invoke()

	return type.fromJsonTree(this[key])
}

fun JsonObject.getArray(key: String): JsonArray {
	if (!has(key)) throw JsonSyntaxException("Expected array at $key, got nothing")
	return this[key] as? JsonArray ?: throw JsonSyntaxException("Expected array at $key, got ${this[key]}")
}

fun JsonObject.getObject(key: String): JsonObject {
	if (!has(key)) throw JsonSyntaxException("Expected object at $key, got nothing")
	return this[key] as? JsonObject ?: throw JsonSyntaxException("Expected object at $key, got ${this[key]}")
}
